import asyncio
import logging

import aioboto3

from botocore.exceptions import ClientError

from .dynamodb import DynamoDB, CheckpointTimeoutException
from .boto_exceptions import RETRY_EXCEPTIONS

log = logging.getLogger(__name__)


class ShardClosedException(Exception):
    pass


class RetryGetRecordsException(Exception):
    pass


class StoppableProcess:

    def __init__(self):
        self.sleep_task = None
        self._stop = False

    def stop(self):
        self._stop = True
        if self.sleep_task and not self.sleep_task.done():
            self.sleep_task.cancel()

    async def interruptable_sleep(self, time):
        if self._stop:
            return False
        self.sleep_task = asyncio.ensure_future(asyncio.sleep(time))
        try:
            await self.sleep_task
            return True
        except asyncio.CancelledError:
            return not self._stop


class AsyncShardReader(StoppableProcess):
    """
    Read from a specific shard, yielding records from async generator
    Remark: how long we sleep between calls to get_records -
            this follow these best practices: http://docs.aws.amazon.com/streams/latest/dev/kinesis-low-latency.html
    """

    def __init__(self, shard_id, shard_iter, dynamodb, checkpoint_interval, sleep_time, consumer):

        super(AsyncShardReader, self).__init__()

        self.shard_id = shard_id
        self.shard_iter = shard_iter
        self.sleep_time = sleep_time
        self.checkpoint_interval = checkpoint_interval
        self.dynamodb = None
        self.consumer = consumer

        log.debug('Shard reader for %s starting', self.shard_id)

        if dynamodb:
            log.debug('Using checkpoint table %s, checkpointing after every %d-th record',
                      dynamodb, self.checkpoint_interval)
            self.dynamodb = dynamodb

        self.client = consumer._get_kinesis_client()
        self.retries = 0
        self.record_count = 0
        self.is_running = True
        self.millis_behind_latest = 0

    async def _get_records(self):
        """
        Internal get_records method. Tries to read from kinesis shard, will raise
        exception if retrying is needed or shard is closed
        :return: list of records as returned by kinesis client
        """
        if self.shard_iter is None:
            log.debug("Shard %s has been closed, exiting", self.shard_id)
            raise ShardClosedException
        try:
            resp = await self.client.get_records(ShardIterator=self.shard_iter)
        except ClientError as e:
            if e.response.get('Error',{}).get('Code') in RETRY_EXCEPTIONS:
                raise RetryGetRecordsException
            else:
                log.error("Client error occurred while reading: %s", e)
                self.is_running = False
                raise e
        else:
            self.shard_iter = resp.get('NextShardIterator')
            self.millis_behind_latest = resp.get('MillisBehindLatest')
            return resp.get('Records')

    async def get_records(self):
        """
        Async generator for shard records
        Handles timeouts, checkpoints and skips empty records
        :return: list of records as returned by kinesis client (i.e. base64-encoded)
        """

        while True:
            try:
                if self._stop:
                    return
                records = await self._get_records()
                if len(records) > 0:
                    yield records
                # FIXME: Could there be empty records in the list? If yes, should we filter them out?
                self.record_count += len(records)
                if self.dynamodb and self.record_count > self.checkpoint_interval:
                    await self.dynamodb.checkpoint(seq=records[-1]['SequenceNumber'])
                    self.record_count = 0
                self.retries = 0
            except RetryGetRecordsException as e:
                sleep_time = min((
                        30,
                        (self.retries or 1) * self.sleep_time
                    ))
                if self.retries > 5:
                    log.debug("Retrying get_records (#%d %ds): %s", self.retries + 1, sleep_time, e)
                if not await self.interruptable_sleep(sleep_time):
                    return
                self.retries += 1
            except Exception as e:
                self.is_running = False
                # Reporting exit on exceptions
                self.consumer.reader_exited(self.shard_id)
                raise e


class AsyncKinesisConsumer(StoppableProcess):
    """
    Generates a list of AsyncShardReaders adding new reader if resharding happens or reader exits
    """

    DEFAULT_SLEEP_TIME = 0.3
    DEFAULT_CHECKPOINT_INTERVAL = 100
    DEFAULT_LOCK_DURATION = 30

    def __init__(self, stream_name, checkpoint_table=None, host_key=None):
        """
        Initialize Async Kinseis Consumer
        :param stream_name:         stream name to read from
        :param checkpoint_table:    DynamoDB table for checkpointing; If not set, checkpointing is not used
        :param host_key:            Key to identify reader instance; If not set, defaults to FQDN.
        """

        super(AsyncKinesisConsumer, self).__init__()

        self.stream_name = stream_name
        self.kinesis_client = aioboto3.client('kinesis')

        self.checkpoint_table = checkpoint_table
        self.host_key = host_key

        self.shard_readers = {}
        self.dynamodb_instances = {}
        self.stream_data = None
        self.force_rescan = True

        self.checkpoint_interval = AsyncKinesisConsumer.DEFAULT_CHECKPOINT_INTERVAL
        self.lock_duration = AsyncKinesisConsumer.DEFAULT_LOCK_DURATION
        self.reader_sleep_time = AsyncKinesisConsumer.DEFAULT_SLEEP_TIME

    def set_checkpoint_interval(self, interval):
        self.checkpoint_interval = interval

    def set_lock_duraion(self, lock_duration):
        self.lock_duration = lock_duration

    def set_reader_sleep_time(self, sleep_time):
        self.reader_sleep_time = sleep_time

    def _get_kinesis_client(self):
        return self.kinesis_client

    def stop(self):
        for shard_id, reader in self.shard_readers.items():
            log.debug('Stopping reader for shard %s', shard_id)
            reader.stop()
        super().stop()

    def reader_exited(self, shard_id):
        log.debug('Reader for shard %s reported exit', shard_id)
        # Force shard rescan
        self.force_rescan = True
        if self.sleep_task and not self.sleep_task.done():
            self.sleep_task.cancel()

    async def get_shard_readers(self):
        """
        Async generator for shard readers. Yields shard readers for every shard in stream and keeps adding
        new readers in case of re-sharding or if reader exited due to error
        """

        stream_data = {}
        while True:
            # Check if all readers are alive and kicking
            shard_readers = {}
            for shard_id, shard_reader in self.shard_readers.items():
                if not shard_reader.is_running:
                    log.debug('Reader for shard %s is not running anymore, forcing rescan', shard_id)
                    self.force_rescan = True
                else:
                    shard_readers[shard_id] = shard_reader
            self.shard_readers = shard_readers
            if self.force_rescan:
                log.debug("Getting description for stream '%s'", self.stream_name)
                stream_data = await self.kinesis_client.describe_stream(StreamName=self.stream_name)
                # XXX TODO: handle StreamStatus -- our stream might not be ready, or might be deleting

                log.debug('Stream has %d shard(s)', len(stream_data['StreamDescription']['Shards']))
            # locks have to be either acquired first time or refreshed
            for shard_data in stream_data['StreamDescription']['Shards']:
                iterator_args = dict(ShardIteratorType='LATEST')
                shard_id = shard_data['ShardId']
                # see if we can get a lock on this shard id
                dynamodb = None
                if self.checkpoint_table:
                    try:
                        dynamodb = self.dynamodb_instances.get(shard_id,
                            DynamoDB(
                                table_name=self.checkpoint_table,
                                shard_id=shard_data['ShardId'],
                                host_key=self.host_key
                            ))
                        shard_locked = await dynamodb.lock_shard(self.lock_duration)
                        self.dynamodb_instances[shard_id] = dynamodb
                    except CheckpointTimeoutException as e:
                        log.warning('Timeout while locking shard %s: %s', shard_id, e)
                        pass
                    except Exception as e:
                        raise e
                    else:
                        if not shard_locked:
                            if shard_id in self.shard_readers \
                                    and not self.shard_readers[shard_id].is_running:
                                log.warning("Can't acquire lock on shard %s and ShardReader is not running", shard_data['ShardId'])
                            # Since we failed to lock the shard we just continue to the next one
                            # Hopefully the lock will expire soon and we can acquire it on the next try
                            continue
                        else:
                            iterator_args = dynamodb.get_iterator_args()
                if shard_id not in self.shard_readers or not self.shard_readers[shard_id].is_running:

                    log.debug("%s iterator arguments: %s", shard_id, iterator_args)

                    # get our initial iterator
                    shard_iter = await self.kinesis_client.get_shard_iterator(
                        StreamName=self.stream_name,
                        ShardId=shard_id,
                        **iterator_args
                    )
                    # create and yield ShardReader
                    shard_reader = AsyncShardReader(
                        shard_id=shard_id,
                        shard_iter=shard_iter.get('ShardIterator'),
                        dynamodb=dynamodb,
                        checkpoint_interval=self.checkpoint_interval,
                        sleep_time=self.reader_sleep_time,
                        consumer=self
                    )
                    self.shard_readers[shard_id] = shard_reader
                    yield shard_reader

            self.force_rescan = False
            # Sleep and check if re-sharding happened
            # If interruptable_sleep returned false, we were signalled to stop
            if not await self.interruptable_sleep(self.lock_duration * 0.8):
                return




import logging
import time
import aioboto3

log = logging.getLogger(__name__)


class AsyncKinesisProducer:

    def __init__(self, stream_name, ordered=True):

        self.stream_name = stream_name
        self.ordered = ordered

        self.seq = '0'

        self.kinesis_client = aioboto3.client('kinesis')
        log.debug("Configured kinesis producer for stream '%s'; ordered=%s",
                  stream_name, ordered)

    async def put_record(self, record, partition_key=None, explicit_hash_key=None):

        if partition_key is None:
            partition_key = '{0}{1}'.format(time.process_time(), time.time())

        kwargs = {
            'StreamName': self.stream_name,
            'Data': record,
            'PartitionKey': partition_key,
        }

        if self.ordered:
            kwargs['SequenceNumberForOrdering'] = self.seq

        if explicit_hash_key:
            kwargs['ExplicitHashKey'] = explicit_hash_key

        resp = await self.kinesis_client.put_record(**kwargs)
        if self.ordered:
            self.seq = resp.get('SequenceNumber')

    # TODO: Add bulk put_records


import asyncio
import logging
import socket
import time

import aioboto3

from botocore.exceptions import ClientError

from .boto_exceptions import RETRY_EXCEPTIONS

log = logging.getLogger(__name__)


class CheckpointTimeoutException(Exception):
    pass


class DynamoDB:
    """
    Class for checkpointing stream using async calls to DynamoDB
    Basically, state.py from kinesis-python client by Evan Borgstrom <eborgstrom@nerdwallet.com>,
    but with async calls and tailored for per-shard operations
    """

    DEFAULT_MAX_RETRIES = 5

    def __init__(self, table_name, shard_id, max_retries=DEFAULT_MAX_RETRIES, host_key=None):
        """
        Initalize DynamoDB
        :param table_name:  DynamoDB table name
        :param shard_id:    ShardId as returned by kinesis client
        :param max_retries: Max retries for communicating with DynamoDB
        :param host_key:    Key to identify reader. If no key provided, defaults to FQDN
        """
        self.table_name = table_name
        self.shard_id = shard_id
        self.shard = {}
        self.retries = 0
        self.max_retries = max_retries
        self.retry_sleep_time = 1
        self.host_key = host_key or socket.getfqdn()
        self.dynamo_table = aioboto3.resource('dynamodb').Table(self.table_name)

    def get_iterator_args(self):
        if self.shard is not None and self.shard.get('seq') is not None:
            return dict(
                ShardIteratorType='AFTER_SEQUENCE_NUMBER',
                StartingSequenceNumber=self.shard.get('seq')
            )
        else:
            return dict(
                ShardIteratorType='LATEST'
            )

    async def checkpoint(self, seq):
        """
        Write checkpoint, so we know where in the stream we've been
        :param seq: kinesis sequence number
        """
        retries = self.max_retries
        while retries > 0:
            try:
                # update the seq attr in our item
                # ensure our fqdn still holds the lock and the new seq is bigger than what's already there
                await self.dynamo_table.update_item(
                    Key={'shard': self.shard_id},
                    UpdateExpression="set seq = :seq",
                    ConditionExpression="fqdn = :fqdn AND (attribute_not_exists(seq) OR seq < :seq)",
                    ExpressionAttributeValues={
                        ':fqdn': self.host_key,
                        ':seq': seq,
                    }
                )
                log.debug('Shard %s: checkpointed seq %s', self.shard_id, seq)
                return
            except ClientError as exc:
                if exc.response('Error', {}).get('Code') in RETRY_EXCEPTIONS:
                    log.warning("Throttled while trying to read lock table in Dynamo: %s", exc)
                    await asyncio.sleep(self.retry_sleep_time)
                    retries -= 1
                    continue
                else:
                    # for all other exceptions (including condition check failures) we just re-raise
                    raise exc

    async def lock_shard(self, lock_holding_time):
        """
        Lock shard, so no other instance will use it
        :param lock_holding_time: how long to hold the lock
        :return: True if lock was obtained, False otherwise
        """

        log.debug('Trying to lock shard %s', self.shard_id)
        dynamo_key = {'shard': self.shard_id}
        fqdn = self.host_key
        now = time.time()
        expire_time = int(now + lock_holding_time)  # dynamo doesn't support floats

        retries = self.max_retries
        while retries > 0:
            try:
                # Do a consistent read to get the current document for our shard id
                resp = await self.dynamo_table.get_item(Key=dynamo_key, ConsistentRead=True)
                self.shard = resp.get('Item')
                break
            except ClientError as e:
                if e.response.get('Error', {}).get('Code') in RETRY_EXCEPTIONS:
                    log.warning("Throttled while trying to read lock table in DynamoDB: %s", e)
                    await asyncio.sleep(self.retry_sleep_time)
                    retries -= 1
                    continue
                else:
                    # all other client errors just get re-raised
                    raise e

        # if there's no Item in the resp then the document didn't exist
        if self.shard is not None and fqdn != self.shard.get('fqdn') and now < self.shard.get('expires'):
            # we don't hold the lock and it hasn't expired
            log.debug("Not starting reader for shard %s -- locked by %s until %s",
                      self.shard_id, self.shard.get('fqdn'), self.shard.get('expires'))
            return False

        retries = self.max_retries
        while retries > 0:
            try:
                if self.shard is not None:
                    # Try to acquire the lock by setting our fqdn and calculated expires.
                    # We add a condition that ensures the fqdn & expires from the document we loaded hasn't changed to
                    # ensure that someone else hasn't grabbed a lock first.
                    await self.dynamo_table.update_item(
                        Key=dynamo_key,
                        UpdateExpression="set fqdn = :new_fqdn, expires = :new_expires",
                        ConditionExpression="fqdn = :current_fqdn AND expires = :current_expires",
                        ExpressionAttributeValues={
                            ':new_fqdn': fqdn,
                            ':new_expires': expire_time,
                            ':current_fqdn': self.shard.get('fqdn'),
                            ':current_expires': self.shard.get('expires'),
                        }
                    )
                    break
                else:
                    # No previous lock. Here our condition prevents a race condition with two readers
                    # starting up and both adding a lock at the same time.
                    await self.dynamo_table.update_item(
                        Key=dynamo_key,
                        UpdateExpression="set fqdn = :new_fqdn, expires = :new_expires",
                        ConditionExpression="attribute_not_exists(#shard_id)",
                        ExpressionAttributeValues={
                            ':new_fqdn': fqdn,
                            ':new_expires': expire_time,
                        },
                        ExpressionAttributeNames={
                            # 'shard' is a reserved word in expressions so we need to use a bound name to work around it
                            '#shard_id': 'shard',
                        },
                        ReturnValues='ALL_NEW'
                    )
                    break
            except ClientError as e:
                if e.response.get('Error', {}).get('Code') == "ConditionalCheckFailedException":
                    # someone else grabbed the lock first
                    log.debug('Shard %s already locked by another instance', self.shard_id)
                    return False

                if e.response('Error', {}).get('Code') in RETRY_EXCEPTIONS:
                    log.warning('Throttled while trying to write lock table in Dynamo: %s', e)
                    await asyncio.sleep(self.retry_sleep_time)
                    retries -= 1
                else:
                    raise e

        # we now hold the lock
        log.debug('Locked shard %s for host %s', self.shard_id, fqdn)
        return True
